from datetime import datetime
from enum import Enum
from typing import List, Literal, NamedTuple, Optional

from pydantic import BaseModel, Extra, Field, NonNegativeFloat, root_validator, PrivateAttr, validator

from chaiverse.schemas.date_range_schema import DateRange, BoundedDateRange


class CompetitionChatRouting(BaseModel):
    percentage: NonNegativeFloat = Field(default=100)

    @validator('percentage')
    def validate_percentage(cls, value):
        assert value <= 100, 'percentage more than 100'
        return value


class Competition(BaseModel, extra=Extra.allow):
    competition_id: str
    leaderboard_format: Optional[str]
    leaderboard_should_use_feedback: Optional[bool]
    submission_date_range: Optional[DateRange]
    evaluation_date_range: Optional[BoundedDateRange]
    chat_routing: CompetitionChatRouting = Field(default=CompetitionChatRouting())
    enrolled_submission_ids: Optional[List[str]]
    created_at: Optional[datetime]

    @root_validator
    def validate_date_ranges(cls, values):
        submission_date_range = values.get('submission_date_range')
        evaluation_date_range = values.get('evaluation_date_range')
        if evaluation_date_range:
            assert submission_date_range, 'missing submission_date_rage'
            assert evaluation_date_range.start_epoch_time >= submission_date_range.end_epoch_time, 'evaluation starts before submission completes'
        return values



import os


DEFAULT_MAX_WORKERS = 1


PUBLIC_LEADERBOARD_MINIMUM_FEEDBACK_COUNT = 0


MODEL_EVAL_SCORE_COLS = ['stay_in_character', 'user_preference', 'entertaining']


LEADERBOARD_FORMAT_CONFIGS = {}


LEADERBOARD_FORMAT_CONFIGS['meval'] = {
    "output_columns": [
        'developer_uid',
        'model_name',
        'is_custom_reward',
        'stay_in_character',
        'user_preference',
        'entertaining',
        'safety_score',
        "overall_rank",
        'elo_rating',
        'num_battles',
        'num_wins',
        'thumbs_up_ratio',
        'double_thumbs_up_ratio',
        'size',
        'status',
        'submission_id',
    ],
    "sort_params": {
        "by": "overall_score",
        "ascending": True
    }
}


LEADERBOARD_FORMAT_CONFIGS['thumbs_up'] = {
    "output_columns": [
        'developer_uid',
        'model_name',
        'thumbs_up_ratio',
        'double_thumbs_up_ratio',
        'overall_rank',
        'total_feedback_count',
        'repetition',
        'stay_in_character',
        'user_preference',
        'entertaining',
        'safety_score',
        'is_custom_reward',
        'submission_id',
        'size',
    ],
    "sort_params": {
        "by": "thumbs_up_ratio",
        "ascending": False
    }
}



__all__ = ["get_leaderboard"]


import numpy as np
import pandas as pd

from chaiverse.utils import get_submissions, distribute_to_workers
from chaiverse.metrics.feedback_metrics import get_metrics_from_feedback
from chaiverse.metrics.submission_metrics import get_metrics_from_submission
from chaiverse import constants, feedback


def get_leaderboard(
        developer_key=None,
        max_workers=constants.DEFAULT_MAX_WORKERS,
        submission_date_range=None,
        evaluation_date_range=None,
        submission_ids=None,
        fetch_feedback=False,
        ):
    submissions = get_submissions(developer_key, submission_date_range)
    submissions = _filter_submissions_by_submission_ids(submissions, submission_ids) if submission_ids != None else submissions
    submissions = _filter_submissions_by_feedback_count(submissions, constants.PUBLIC_LEADERBOARD_MINIMUM_FEEDBACK_COUNT)
    df = distribute_to_workers(
        get_leaderboard_row,
        submissions.items(),
        developer_key=developer_key,
        evaluation_date_range=evaluation_date_range,
        max_workers=max_workers,
        fetch_feedback=fetch_feedback
    )
    df = pd.DataFrame(df)
    if len(df):
        df = _get_filled_leaderboard(df)
        df.index = np.arange(1, len(df)+1)
    return df


def get_leaderboard_row(submission_item, developer_key=None, evaluation_date_range=None, fetch_feedback=False):
    submission_id, submission_data = submission_item
    if fetch_feedback:
        submission_feedback_total = submission_data.get('double_thumbs_up', 0) + submission_data['thumbs_up'] + submission_data['thumbs_down']
        is_updated = feedback.is_submission_updated(submission_id, submission_feedback_total)
        metrics = get_metrics_from_feedback(
            submission_id, 
            developer_key, 
            reload=is_updated, 
            evaluation_date_range=evaluation_date_range
        )
    else:
        metrics = get_metrics_from_submission(submission_data) 
    return {'submission_id': submission_id, **submission_data, **metrics}


def _get_filled_leaderboard(df):
    # maintain backwards compatibility with model_name field
    _fill_default_value(df, 'model_name', df['submission_id'])
    _fill_default_value(df, 'is_custom_reward', False)
    for col in _get_filled_columns():
        _fill_default_value(df, col, None)
    return df


def _get_filled_columns():
    columns = []
    for leaderboard_format in constants.LEADERBOARD_FORMAT_CONFIGS.keys():
        new_columns = constants.LEADERBOARD_FORMAT_CONFIGS[leaderboard_format]['output_columns']
        columns.extend(new_columns)
    return columns


def _fill_default_value(df, field, default_value):
    if field not in df:
        df[field] = None
    if default_value is not None:
        df[field] = df[field].fillna(default_value)


def _filter_submissions_by_submission_ids(submissions, submission_ids):
    filtered_submissions = {
        submission_id: data
        for submission_id, data in submissions.items()
        if submission_id in submission_ids
    }
    return filtered_submissions


def _filter_submissions_by_feedback_count(submissions, min_feedback_count):
    submissions = {
        submission_id: submission_data for submission_id, submission_data in submissions.items()
        if submission_data.get('double_thumbs_up', 0) + submission_data['thumbs_up'] + submission_data['thumbs_down'] >= min_feedback_count
    }
    return submissions



__all__ = ["ConversationMetrics"]


import string

import numpy as np

class ConversationMetrics():
    def __init__(self, messages):
        self.messages = messages

    @property
    def mcl(self):
        return len([m for m in self.messages if not m['deleted']])

    @property
    def repetition_score(self):
        responses = [m['content'] for m in self.messages if not self._is_from_user(m)]
        score = np.nan if len(responses) < 2 else get_repetition_score(responses)
        return score

    def _is_from_user(self, message):
        return '_bot' not in message['sender']['uid']


def get_repetition_score(responses):
    # average jaccard similarities over unigrams
    list_of_tokens = _tokenize_responses(responses)
    pairs = zip(list_of_tokens[:-1], list_of_tokens[1:])
    similarities = [_get_jaccard_similarity(set1, set2) for set1, set2 in pairs]
    return np.mean(similarities)


def _get_jaccard_similarity(set1, set2):
    intersection_len = len(set1.intersection(set2))
    union_len = len(set1.union(set2))
    return intersection_len / union_len


def _tokenize_responses(responses):
    return [set(_remove_punctuation(text).split()) for text in responses]


def _remove_punctuation(text):
    translation_table = str.maketrans('', '', string.punctuation)
    cleaned_text = text.translate(translation_table)
    if len(cleaned_text.split()) == 0:
        cleaned_text = '...'
    return cleaned_text.lower()



from chaiverse.lib import binomial_tools


def get_metrics_from_submission(submission):
    double_thumbs_up =  submission.get('double_thumbs_up', 0)
    thumbs_up = submission['thumbs_up']
    thumbs_down = submission['thumbs_down']
    submission_feedback_total = double_thumbs_up + thumbs_up + thumbs_down

    metrics = {
        'thumbs_up_ratio': binomial_tools.get_ratio(thumbs_up, thumbs_down + double_thumbs_up),
        'thumbs_up_ratio_se': binomial_tools.get_ratio_se(thumbs_up, thumbs_down + double_thumbs_up),
        'double_thumbs_up_ratio': binomial_tools.get_ratio(double_thumbs_up, thumbs_down + thumbs_up),
        'double_thumbs_up_ratio_se': binomial_tools.get_ratio_se(double_thumbs_up, thumbs_down + thumbs_up),
        'total_feedback_count': submission_feedback_total,
    }
    return metrics


from chaiverse.http_client import SubmitterClient
from chaiverse.login_cli import auto_authenticate
from chaiverse.schemas import Competition
from chaiverse import config


@auto_authenticate
def get_competitions(developer_key):
    submitter_client = SubmitterClient(developer_key=developer_key)
    competitions = submitter_client.get(config.COMPETITIONS_ENDPOINT)
    competitions = [Competition(**competition) for competition in competitions]
    return competitions



BASE_SUBMITTER_URL = "https://guanaco-submitter.chai-research.com"
BASE_FEEDBACK_URL = "https://guanaco-feedback.chai-research.com"

LEADERBOARD_ENDPOINT = "/leaderboard"
CHAT_ENDPOINT = "/models/{submission_id}/chat"
FEEDBACK_ENDPOINT = "/feedback/{submission_id}"

SUBMISSION_ENDPOINT = "/models/submit"
BLEND_SUBMISSION_ENDPOINT = "/models/submit_blend"
ALL_SUBMISSION_STATUS_ENDPOINT = "/models/"
INFO_ENDPOINT = "/models/{submission_id}"
DEACTIVATE_ENDPOINT = "/models/{submission_id}/deactivate"
REDEPLOY_ENDPOINT = "/models/{submission_id}/redeploy"
EVALUATE_ENDPOINT = "/models/{submission_id}/evaluate"
TEARDOWN_ENDPOINT = "/models/{submission_id}/teardown"

COMPETITIONS_ENDPOINT = '/competitions'
PENDING_COMPETITION_ENDPOINT = '/pending_competition'
ACTIVE_COMPETITION_ENDPOINT = '/active_competition'
